use hmac::{Hmac, Mac};
use sha2::{Digest, Sha256};

pub struct CloudSecurity {
    login_key: &'static str,
    iot_key: &'static str,
    hmac_key: &'static str,

    access_token: Option<String>,
    random_data: Option<String>,
}

impl CloudSecurity {
    pub fn new(login_key: &'static str, iot_key: &'static str, hmac_key: &'static str) -> Self {
        Self {
            login_key,
            iot_key,
            hmac_key,

            access_token: None,
            random_data: None,
        }
    }

    pub fn set_aes_keys(&mut self, access_token: &str, random_data: &str) {
        self.access_token = Some(access_token.to_string());
        self.random_data = Some(random_data.to_string());
    }

    pub fn sign(&self, data: &str, random: &str) -> String {
        let mut msg = self.iot_key.to_string();

        msg += data;
        msg += random;

        type HmacSha256 = Hmac<Sha256>;

        let mut mac = HmacSha256::new_from_slice(self.hmac_key.as_bytes()).unwrap();
        mac.update(msg.as_bytes());

        mac.finalize()
            .into_bytes()
            .iter()
            .map(|b| format!("{b:02x}"))
            .collect::<Vec<String>>()
            .join("")
    }

    pub fn device_id(account: &str) -> String {
        let mut sha = Sha256::digest(format!("Hello, {account}!").as_bytes())
            .iter()
            .map(|b| format!("{b:02x}"))
            .collect::<Vec<String>>()
            .join("");

        sha.truncate(16);

        sha
    }

    pub fn encrypt_iam_password(&self, login_id: &str, data: &str) -> String {
        let md = md5::compute(data.as_bytes());
        let md_hex = format!("{:x}", md);

        let md_second = md5::compute(md_hex.as_bytes());
        let md_second_hex = format!("{:x}", md_second);

        let login_hash = format!("{}{}{}", login_id, md_second_hex, self.login_key);

        Sha256::digest(login_hash.as_bytes())
            .iter()
            .map(|b| format!("{b:02x}"))
            .collect::<Vec<String>>()
            .join("")
    }

    pub fn encrypt_password(&self, login_id: &str, data: &str) -> String {
        let m_hex = Sha256::digest(data.as_bytes())
            .iter()
            .map(|b| format!("{b:02x}"))
            .collect::<Vec<String>>()
            .join("");

        let login_hash = format!("{}{}{}", login_id, m_hex, self.login_key);

        Sha256::digest(login_hash.as_bytes())
            .iter()
            .map(|b| format!("{b:02x}"))
            .collect::<Vec<String>>()
            .join("")
    }

    pub fn udp_id(device_id: u64, method: u8) -> String {
        let bytes_id: Vec<u8> = match method {
            0 => device_id.to_be_bytes().into_iter().rev().collect(),
            1 => {
                let mut v = device_id.to_be_bytes().to_vec();
                v.truncate(6);

                v
            }
            2 => {
                let mut v = device_id.to_le_bytes().to_vec();
                v.truncate(6);

                v
            }

            _ => return String::new(),
        };

        let mut data = Sha256::digest(&bytes_id);

        for i in 0..16 {
            data[i] ^= data[i + 16];
        }

        let mut v: Vec<u8> = data.into_iter().collect();
        v.truncate(16);

        v.iter()
            .map(|b| format!("{b:02x}"))
            .collect::<Vec<String>>()
            .join("")
    }
}

#[cfg(test)]
mod test {
    use crate::cloud::Cloud;

    use super::*;

    #[test]
    fn iam_password() {
        let sec = CloudSecurity::new(Cloud::APP_KEY, "", "");

        let sha = sec.encrypt_iam_password("bd3937c1-49ba-418c-a6e0-4b600263", "Hoda.semi1");

        debug_assert_eq!(
            sha,
            "6cae02a0e1bc0c05ef886b1e045fc0ae662d9bc3ac60b890fa1c4bd59699c944"
        );
    }

    #[test]
    fn password() {
        let sec = CloudSecurity::new(Cloud::APP_KEY, "", "");

        let enc = sec.encrypt_password("bd3937c1-49ba-418c-a6e0-4b600263", "Hoda.semi1");

        debug_assert_eq!(
            enc,
            "ac57663c18c81ad0423edb235d5b1059d792c2a18d5fad70a83a4afa92affabb"
        );
    }
}
